from lerobot.common.robot_devices.motors.configs import FeetechMotorsBusConfig
from lerobot.common.robot_devices.motors.feetech import FeetechMotorsBus, TorqueMode
from lerobot.common.robot_devices.robots.utils import make_robot_config

from slobot.configuration import Configuration

import json
import numpy as np

class Feetech():
    ROBOT_TYPE = 'so100'
    ARM_NAME = 'main'
    ARM_TYPE = 'follower'

    DOF = 6
    MODEL_RESOLUTION = 4096
    MOTOR_DIRECTION = [-1, 1, 1, 1, 1, 1]

    PORT = '/dev/ttyACM0'

    def calibrate_pos(preset):
        feetech = Feetech()
        feetech.calibrate(preset)

    def move_to_pos(pos):
        feetech = Feetech()
        feetech.move(pos)

    def __init__(self, **kwargs):
        self.motors_bus = self._create_motors_bus()
        self.qpos_handler = kwargs.get('qpos_handler', None)

    def disconnect(self):
        self.set_torque(False)
        self.motors_bus.disconnect()

    def get_qpos(self):
        return self.pos_to_qpos(self.get_pos())

    def get_pos(self):
        return self.motors_bus.read('Present_Position')

    def handle_qpos(self, qpos):
        pos = self.qpos_to_pos(qpos)
        self.set_pos(pos)

    def qpos_to_pos(self, qpos):
        pos = [ self._qpos_to_steps(qpos, i)
            for i in range(Feetech.DOF) ]
        return pos

    def pos_to_qpos(self, pos):
        qpos = [ self._steps_to_qpos(pos, i)
            for i in range(Feetech.DOF) ]
        return qpos

    def set_pos(self, pos):
        self.set_torque(True)
        self.motors_bus.write('Goal_Position', pos)
        if self.qpos_handler is not None:
            qpos = self.pos_to_qpos(pos)
            print("feetech qpos", qpos)
            self.qpos_handler.handle_qpos(qpos)
    
    def set_torque(self, is_enabled):
        torque_enable = TorqueMode.ENABLED.value if is_enabled else TorqueMode.DISABLED.value 
        self.motors_bus.write('Torque_Enable', torque_enable)

    def move(self, target_pos):
        self.set_pos(target_pos)
        position = self.get_pos()
        error = np.linalg.norm(target_pos - position) / Feetech.MODEL_RESOLUTION
        print("pos error=", error)

    def calibrate(self, preset):
        input(f"Move the arm to the {preset} position ...")
        pos = self.get_pos()
        pos_json = json.dumps(pos.tolist())
        print(f"Current position is {pos_json}")

    def _create_motors_bus(self):
        self.robot_config = make_robot_config(Feetech.ROBOT_TYPE)
        motors = self.robot_config.follower_arms[Feetech.ARM_NAME].motors
        config = FeetechMotorsBusConfig(port=self.PORT, motors=motors)
        motors_bus = FeetechMotorsBus(config)
        motors_bus.connect()
        return motors_bus

    def _qpos_to_steps(self, qpos, motor_index):
        steps = Feetech.MOTOR_DIRECTION[motor_index] * Feetech.MODEL_RESOLUTION * (qpos[motor_index] - Configuration.QPOS_MAP['rotated'][motor_index]) / (2 * np.pi)
        return Configuration.POS_MAP['rotated'][motor_index] + int(steps)

    def _steps_to_qpos(self, pos, motor_index):
        steps = pos[motor_index] - Configuration.POS_MAP['rotated'][motor_index]
        return Configuration.QPOS_MAP['rotated'][motor_index] + Feetech.MOTOR_DIRECTION[motor_index] * steps * (2 * np.pi) / Feetech.MODEL_RESOLUTION


import time
import os
import threading
import queue
import numpy as np
import matplotlib.pyplot as plt

from moviepy.video.io.ffmpeg_writer import FFMPEG_VideoWriter
from slobot.configuration import Configuration
from slobot.so_arm_100 import SoArm100
from slobot.simulation_frame import SimulationFrame
from slobot.simulation_frame_paths import SimulationFramePaths
from slobot.video_writer import VideoWriter


class VideoStreams:
    FRAME_TYPES = ["rgb", "depth", "segmentation", "normal"]

    def __init__(self, **kwargs):
        os.makedirs(Configuration.WORK_DIR, exist_ok=True)
        self.logger = Configuration.logger(__name__)

        self.video_segment_queue = queue.Queue()

        self.codec = kwargs.get('codec', 'libx264')

    def frame_filenames(self, res, fps, segment_duration):
        # run simulation in a separate thread
        thread = threading.Thread(target=self.run_simulation, args=(res, fps, segment_duration))
        thread.start()

        while True:
            simulation_frame_paths = self.video_segment_queue.get()
            if simulation_frame_paths is None:
                break

            yield simulation_frame_paths

        thread.join()

    def run_simulation(self, res, fps, segment_duration):
        cam_id = 0
        env_id = 0
        self.start(cam_id, [env_id], res, fps, segment_duration, rgb=True, depth=True, segmentation=True, normal=True)

        mjcf_path = Configuration.MJCF_CONFIG
        arm = SoArm100(mjcf_path=mjcf_path, frame_handler=self, res=res, show_viewer=False)
        arm.elemental_rotations()

        self.stop()

        # stop genesis
        arm.genesis.stop()

    def start(
        self,
        cam_id,
        env_ids,
        res,
        fps,
        segment_duration,
        rgb=True,
        depth=False,
        segmentation=False,
        normal=False
    ):
        self.cam_id = cam_id
        self.env_ids = env_ids
        self.res = res
        self.fps = fps

        self.period = 1.0 / self.fps
        self.segment_id = 0

        self.frame_enabled = [rgb, depth, segmentation, normal]

        self.video_timestamp = time.time()
        self.current_frame = None

        self.segment_duration = segment_duration

        self.simulation_frames = []

    def handle_frame(self, frame):
        timestamp = time.time()

        rgb_arr, depth_arr, seg_arr, normal_arr = frame

        if depth_arr is not None:
            depth_arr = VideoStreams.logarithmic_depth_to_rgb(depth_arr)

        simulation_frame = SimulationFrame(timestamp, rgb_arr, depth_arr, seg_arr, normal_arr)

        self.simulation_frames.append(simulation_frame)

        if self.video_timestamp + self.segment_duration <= simulation_frame.timestamp:
            self.do_transcode(self.simulation_frames)

    def do_transcode(self, simulation_frames):
        if len(simulation_frames) == 0:
            return

        simulation_frame_paths = self.transcode_frames(simulation_frames)
        self.video_segment_queue.put(simulation_frame_paths)

    def transcode_frames(self, simulation_frames) -> SimulationFramePaths:
        self.logger.info(f"Transcoding video segment {self.segment_id}")
        date_time = time.strftime('%Y%m%d_%H%M%S')

        if self.current_frame is None:
            self.current_frame = simulation_frames[0]

        first_timestamp = simulation_frames[0].timestamp

        video_frames = []
        for simulation_frame in simulation_frames:
            while self.video_timestamp + self.period < simulation_frame.timestamp:
                video_frames.append(self.current_frame)
                self.video_timestamp += self.period

            # TODO last frame of the simulation will be dropped
            self.current_frame = simulation_frame

        simulation_frames.clear()

        simulation_frame_videos = []
        for env_id in self.env_ids:
            env_simulation_frame_videos = []
            for frame_id in range(len(self.FRAME_TYPES)):
                if not self.frame_enabled[frame_id]:
                    next

                filename = self._filename(self.cam_id, env_id, self.FRAME_TYPES[frame_id], date_time, self.segment_id)

                video_writer = VideoWriter(filename, self.res, self.fps, codec=self.codec)
                type_frames = [
                    simulation_frame.frame(frame_id)
                    for simulation_frame in video_frames
                ]
                if len(self.env_ids) > 1:
                    type_frames = [
                        env_frame[env_id]
                        for env_frame in type_frames
                    ]
                type_frames = np.array(type_frames)
                video_writer.transcode(type_frames, filename)

                env_simulation_frame_videos.append(filename)
            simulation_frame_videos.append(env_simulation_frame_videos)

        self.logger.info(f"Done flushing frames for segment {self.segment_id}")
        self.segment_id += 1

        return SimulationFramePaths(first_timestamp, simulation_frame_videos)

    def stop(self):
        self.do_transcode(self.simulation_frames)

        self.video_segment_queue.put(None) # add poison pill

    def _filename(self, cam_id, env_id, frame_type, date_time, segment_id):
        return f"{Configuration.WORK_DIR}/cam_{cam_id}_env_{env_id}_{frame_type}_{date_time}_{segment_id}.ts"

    def logarithmic_depth_to_rgb(depth_arr):
        """
        Use logarithmic scaling to enhance depth visualization
        Helps spread out colors better than linearly potentially improving contrast
        """
        # Add small epsilon to avoid log(0)
        log_depth = np.log1p(depth_arr - np.min(depth_arr))
        normalized_log_depth = (log_depth - np.min(log_depth)) / (np.max(log_depth) - np.min(log_depth))

        # Use a perceptually uniform colormap for better distinction
        depth_rgb = plt.cm.plasma(normalized_log_depth) * 255
        depth_rgb = depth_rgb.astype(np.uint8)
        return depth_rgb[:, :, :3]


import gradio as gr
from slobot.image_streams import ImageStreams

class GradioImageApp():
    def __init__(self):
        self.image_streams = ImageStreams()

    def launch(self):
        with gr.Blocks() as demo:
            with gr.Row():
                button = gr.Button()
                width = gr.Number(label='Width', value=640)
                height = gr.Number(label='Height', value=480)
                fps = gr.Slider(label='FPS', minimum=1, maximum=10, value=3, step=1)
            with gr.Row():
                rgb = gr.Image(label='RGB')
                depth = gr.Image(label='Depth')
            with gr.Row():
                segmentation = gr.Image(label='Segmentation Mask')
                surface = gr.Image(label='Surface Normal')

            button.click(self.sim_images, [width, height, fps], [rgb, depth, segmentation, surface])

        demo.launch()

    def sim_images(self, width, height, fps):
        res = (width, height)
        for simulation_frame_paths in self.image_streams.frame_filenames(res, fps):
            yield simulation_frame_paths.paths



import numpy as np
import torch

from slobot.genesis import Genesis
from slobot.configuration import Configuration

class SoArm100():
    JAW_LENGTH = 0.107
    JAW_WIDTH = 0.008
    JAW_DEPTH = 0.0005

    # Mujoco home position
    HOME_QPOS = [0, -np.pi/2, np.pi/2, np.pi/2, -np.pi/2, 0]

    def sim_qpos(target_qpos):
        mjcf_path = Configuration.MJCF_CONFIG
        arm = SoArm100(mjcf_path=mjcf_path)
        arm.genesis.entity.set_qpos(target_qpos)
        arm.genesis.hold_entity()

    def __init__(self, **kwargs):
        self.qpos_handler = kwargs.get('qpos_handler', None)

        # provide a callback for each step update of the simulation
        kwargs['step_handler'] = self
        self.genesis = Genesis(**kwargs)

        self.frame_handler = kwargs.get('frame_handler', None)

        self.camera = self.genesis.camera
        self.entity = self.genesis.entity
        self.fixed_jaw = self.genesis.fixed_jaw

    def elemental_rotations(self):
        self.go_home()
        pos = self.fixed_jaw.get_pos()
        quat = self.fixed_jaw.get_quat()

        print("pos=", pos)
        print("quat=", quat)

        euler = self.genesis.quat_to_euler(quat)

        print("euler=", euler)

        steps = 2

        # turn the fixed jaw around the global x axis, from vertical to horizontal
        for roll in np.linspace(np.pi/2, 0, steps):
            euler[0] = roll
            quat = self.genesis.euler_to_quat(euler)
            self.genesis.move(self.fixed_jaw, pos, quat)

        # turn the fixed jaw around the global y axis
        for pitch in np.linspace(0, np.pi, steps):
            euler[1] = pitch
            quat = self.genesis.euler_to_quat(euler)
            self.genesis.move(self.fixed_jaw, pos, quat)

        # turn the fixed jaw around the global z axis
        pos = None
        for yaw in np.linspace(0, np.pi/2, steps):
            euler[2] = yaw
            quat = self.genesis.euler_to_quat(euler)
            self.genesis.move(self.fixed_jaw, pos, quat)

        self.stop()

    def stop(self):
        #self.camera.stop_recording(save_to_filename='so_arm_100.mp4')
        pass

    def go_home(self):
        target_qpos = torch.tensor(SoArm100.HOME_QPOS)
        self.genesis.follow_path(target_qpos)

    def draw_fixed_jaw_arrow(self):
        t = [self.JAW_WIDTH, -self.JAW_LENGTH, -self.JAW_DEPTH]
        self.genesis.draw_arrow(self.fixed_jaw, t)

    def open_jaw(self):
        self.genesis.update_qpos(self.jaw, np.pi/2)

    def handle_step(self):
        if self.frame_handler is not None:
            frame = self.camera.render(rgb=True, depth=True, segmentation=True, colorize_seg=True, normal=True)
            self.frame_handler.handle_frame(frame)

        if self.qpos_handler is not None:
            self.qpos_handler.handle_qpos(self.entity.get_qpos())

    def handle_qpos(self, qpos):
        self.entity.set_qpos(qpos)


import time
import json

from slobot.configuration import Configuration
from slobot.so_arm_100 import SoArm100
from slobot.fps_metric import FpsMetric


class FpsGauge():
    def __init__(self, **kwargs):
        self.max_period = kwargs['max_period']
        self.res = kwargs['res']

    def show_fps(self):
        self.previous_time = time.time()
        self.frames = 0

        mjcf_path = Configuration.MJCF_CONFIG
        arm = SoArm100(mjcf_path=mjcf_path, frame_handler=self, res=self.res, show_viewer=False)
        arm.elemental_rotations()

        arm.genesis.stop()

    def handle_frame(self, frame):
        self.frames += 1

        current_time = time.time()

        diff_time = current_time - self.previous_time
        if diff_time > self.max_period:
            fps = self.frames / diff_time
            self.add_metric(current_time, fps)
            self.frames = 0
            self.previous_time = current_time

    def add_metric(self, timestamp, fps):
        fps_metric = FpsMetric(timestamp, fps)
        print("FPS=", fps_metric)


class SimulationFrame():
    def __init__(self, timestamp, rgb, depth, segmentation, surface):
        self.timestamp = timestamp
        self.rgb = rgb
        self.depth = depth
        self.segmentation = segmentation
        self.surface = surface

    def frame(self, frame_id):
        match frame_id:
            case 0:
                return self.rgb
            case 1:
                return self.depth
            case 2:
                return self.segmentation
            case 3:
                return self.surface


import numpy as np
import logging

class Configuration:
    WORK_DIR = "/tmp/slobot"

    MJCF_CONFIG = './trs_so_arm100/so_arm100.xml'

    # 16:9 aspect ratio
    LD = (426, 240)
    SD = (854, 480)
    HD = (1280, 720)
    FHD = (1920, 1080)

    QPOS_MAP = {
        "rotated": [-np.pi/2, -np.pi/2, np.pi/2, np.pi/2, -np.pi/2, np.pi/2],
        "zero": [0, 0, 0, 0, 0, 0],
        "rest": [0.049, -3.62, 3.19, 1.26, -0.17, -0.67]
    }

    POS_MAP = {
        "rotated": [3147, 2061, 2001, 3017, 1121, 3438],
        "zero": [2112, 3069, 996, 2099, 2111, 2088],
        "rest": [2091, 728, 3056, 2812, 2032, 1977]
    }

    def logger(logger_name):
        logger = logging.getLogger(logger_name)
        logger.setLevel(logging.INFO)
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
        logger.addHandler(handler)
        return logger


import torch

import genesis as gs
from genesis.engine.entities import RigidEntity
from genesis.engine.entities.rigid_entity import RigidLink, RigidJoint

from slobot.configuration import Configuration

import pprint

from scipy.spatial.transform import Rotation

class Genesis():
    POSITION_STEPS = 50

    EXTRINSIC_SEQ = 'xyz'

    def __init__(self, **kwargs):
        gs.init(backend=gs.gpu) # gs.cpu

        vis_mode = 'visual' # collision

        res = kwargs.get('res', Configuration.FHD)
        camera_pos = (0.125, -1, 0.5)

        lookat = (0, 0, 0)

        lights = [
            { "type": "directional", "dir": (1, 1, -1), "color": (1.0, 1.0, 1.0), "intensity": 5.0 },
        ]

        self.step_handler = kwargs['step_handler']

        show_viewer = kwargs.get('show_viewer', True)

        self.scene = gs.Scene(
            show_viewer=show_viewer,
            viewer_options = gs.options.ViewerOptions(
                res           = res,
                camera_lookat = lookat,
                camera_pos    = camera_pos,
            ),
            vis_options    = gs.options.VisOptions(
                lights          = lights,
                #show_link_frame = True,
            ),
            show_FPS       = False,
        )

        plane = gs.morphs.Plane()
        self.scene.add_entity(plane, vis_mode=vis_mode)

        arm_morph = self.parse_morph(**kwargs)

        self.entity: RigidEntity = self.scene.add_entity(
            arm_morph,
            vis_mode=vis_mode,
        )

        # TODO errors in non-interactive mode in Docker
        #print("Joints=", pprint.pformat(self.entity.joints))
        #print("Links=", pprint.pformat(self.entity.links))

        self.fixed_jaw: RigidLink = self.entity.get_link('Fixed_Jaw')

        self.jaw: RigidJoint = self.entity.get_joint('Jaw')

        self.camera = self.scene.add_camera(
            res    = res,
            pos    = camera_pos,
            lookat = lookat,
        )

        self.scene.build()

        #self.camera.start_recording()

        print("Limits=", self.entity.get_dofs_limit())


        print("Kp=", self.entity.get_dofs_kp())

        print("Kd=", self.entity.get_dofs_kv())
        print("Force range=", self.entity.get_dofs_force_range())

        qpos = self.entity.get_qpos()
        print("qpos=", qpos)

        print("collisions=", self.entity.detect_collision())

        damping = self.entity.get_dofs_damping()
        print("damping", damping)

    def parse_morph(self, **kwargs):
        mjcf_path = kwargs['mjcf_path']
        if mjcf_path is not None:
            return gs.morphs.MJCF(
                file = mjcf_path,
            )

        urdf_path = kwargs['urdf_path']
        if urdf_path is not None:
            return gs.morphs.URDF(
                file  = urdf_path,
                fixed = True,
                convexify = False,
            )

        raise ValueError(f"Provide either mjcf_path or urdf_path")

    def follow_path(self, target_qpos):
        path = self.entity.plan_path(
            qpos_goal        = target_qpos,
            ignore_collision = True,
        )

        if len(path) == 0:
            return

        for waypoint in path:
            self.entity.control_dofs_position(waypoint)
            self.step()

        # allow more steps to the PD controller to stabilize to the target position
        for _ in range(self.POSITION_STEPS):
            self.step()

        current_error = self.qpos_error(target_qpos)
        print("qpos error=", current_error)

    def stop(self):
        gs.destroy()

    def move(self, link, target_pos, target_quat):
        target_qpos = self.entity.inverse_kinematics(
            link     = link,
            pos      = target_pos,
            quat     = target_quat,
        )

        self.follow_path(target_qpos)
        self.validate_target(self.fixed_jaw, target_pos, target_quat)

    def step(self):
        self.scene.step()
        if self.step_handler is not None:
            self.step_handler.handle_step()

    def hold_entity(self):
        target_qpos = self.entity.get_qpos()
        print("sim qpos", target_qpos)
        while True:
            self.entity.control_dofs_position(target_qpos)
            self.step()

    def get_qpos_idx(self, joint):
        return joint.idx_local - 1  # offset the base joint, which is not part of qpos variables

    def update_qpos(self, joint, qpos):
        target_qpos = self.entity.get_qpos()
        joint_idx = self.get_qpos_idx(joint)
        target_qpos[joint_idx] = qpos
        self.follow_path(target_qpos)

    def validate_target(self, link, target_pos, target_quat):
        self.validate_pos(link, target_pos)
        self.validate_quat(link, target_quat)

    def validate_pos(self, link, target_pos):
        if target_pos is None:
            return

        current_pos = link.get_pos()

        current_pos = current_pos.to(target_pos.device)
        error = torch.norm(current_pos - target_pos)
        print("pos error=", error)

    def validate_quat(self, link, target_quat):
        if target_quat is None:
            return

        current_quat = link.get_quat()

        current_quat = current_quat.to(target_quat.device)
        error = torch.norm(current_quat - target_quat)
        print("quat error=", error)

    def qpos_error(self, target_qpos):
        current_qpos = self.entity.get_qpos()

        # To avoid division by 0, create a target_qpos_denominator variable where 0 are replaced with 1
        target_qpos_denominator = torch.where(target_qpos == 0, torch.tensor(1.0, device=target_qpos.device), target_qpos)

        current_qpos = current_qpos.to(target_qpos.device)
        error = torch.abs((current_qpos - target_qpos) / target_qpos_denominator)
        return torch.norm(error)

    def translate(self, pos, t, euler):
        r = Rotation.from_euler(self.EXTRINSIC_SEQ, euler)
        t = r.apply(t)
        return pos + t

    def quat_to_euler(self, quat):
        quat = quat.cpu()
        return Rotation.from_quat(quat, scalar_first=True).as_euler(seq=self.EXTRINSIC_SEQ)

    def euler_to_quat(self, euler):
        quat = Rotation.from_euler(self.EXTRINSIC_SEQ, euler).as_quat(scalar_first=True)
        return torch.tensor(quat)

    def draw_arrow(self, link, t):
        self.scene.clear_debug_objects()

        link_pos = link.get_pos()
        link_quat = link.get_quat()
        link_euler = self.quat_to_euler(link_quat)
        t_pos = self.translate(link_pos, t, link_euler)
        arrow_vec = t_pos - link_pos
        self.scene.draw_debug_arrow(link_pos, arrow_vec, radius = 0.002, color=(1, 0, 0, 1))


import gradio as gr
from slobot.video_streams import VideoStreams

class GradioVideoApp():
    def __init__(self, **kwargs):
        self.video_streams = VideoStreams(**kwargs)

    def launch(self):
        with gr.Blocks() as demo:
            with gr.Row():
                button = gr.Button()
                width = gr.Number(label='Width', value=640)
                height = gr.Number(label='Height', value=480)
                fps = gr.Slider(label='FPS', minimum=1, maximum=60, value=24, step=1)
                segment_duration = gr.Slider(label='Segment Duration', minimum=1, maximum=20, value=10, step=1)
            with gr.Row():
                rgb = gr.Video(label='RGB', streaming=True)
                depth = gr.Video(label='Depth', streaming=True)
            with gr.Row():
                segmentation = gr.Video(label='Segmentation Mask', streaming=True)
                surface = gr.Video(label='Surface Normal', streaming=True)

            button.click(self.sim_videos, [width, height, fps, segment_duration], [rgb, depth, segmentation, surface])

        demo.launch()

    def sim_videos(self, width, height, fps, segment_duration):
        res = (width, height)
        env_id = 0
        for simulation_frame_paths in self.video_streams.frame_filenames(res, fps, segment_duration):
            yield simulation_frame_paths.paths[env_id]



from multiprocessing import shared_memory
import numpy as np
import subprocess

class VideoWriter():
    FFMPEG_BINARY = "ffmpeg"

    def __init__(self, filename, res, fps, codec):
        self.filename = filename
        self.res = res
        self.fps = fps
        self.codec = codec

    def transcode(self, raw_video, output_filename):
        shared_memory = self.shared_memory(raw_video)
        shared_memory_filename = f"/dev/shm/{shared_memory.name}"

        cmd = [
            VideoWriter.FFMPEG_BINARY,
            "-f", "rawvideo",
            "-vcodec", "rawvideo",
            "-pix_fmt", "rgb24",
            "-s", "%dx%d" % (self.res[0], self.res[1]),
            "-r", "%.02f" % self.fps,
            "-an",
            "-i", shared_memory_filename,
            "-vcodec", self.codec,
            output_filename
        ]

        # execute cmd
        process = subprocess.Popen(cmd, stderr=subprocess.PIPE)
        process.wait()
        if process.returncode != 0:
            error_message = process.stderr.read().decode()
            process.stderr.close()
            print(f"FFmpeg error: {error_message}")
            raise RuntimeError(f"FFmpeg failed with error code {process.returncode}")

        shared_memory.unlink()


    def shared_memory(self, raw_video):
        shm = shared_memory.SharedMemory(create=True, size=raw_video.nbytes)
        shm_raw_video = np.ndarray(raw_video.shape, dtype=raw_video.dtype, buffer=shm.buf)
        shm_raw_video[:] = raw_video[:]
        shm.close()
        return shm
    


import time
import os
import threading
import queue

from PIL import Image

from slobot.configuration import Configuration
from slobot.so_arm_100 import SoArm100
from slobot.video_streams import VideoStreams
from slobot.simulation_frame import SimulationFrame
from slobot.simulation_frame_paths import SimulationFramePaths

# Generate a stream of images from the simulation
class ImageStreams:

    def __init__(self):
        os.makedirs(Configuration.WORK_DIR, exist_ok=True)
        self.queue = queue.Queue()

    def frame_filenames(self, res, fps):
        thread = threading.Thread(target=self.run_simulation, args=(res, fps))
        thread.start()

        while True:
            simulation_frame_paths = self.queue.get()
            if simulation_frame_paths is None:
                break

            yield simulation_frame_paths

        thread.join()

    def run_simulation(self, res, fps):
        self.start(res, fps, rgb=True, depth=True, segmentation=True, normal=True)

        mjcf_path = Configuration.MJCF_CONFIG
        arm = SoArm100(mjcf_path=mjcf_path, frame_handler=self, res=res, show_viewer=False)
        arm.elemental_rotations()

        self.stop()

        # stop genesis
        arm.genesis.stop()

    def start(
        self,
        res,
        fps,
        rgb=True,
        depth=False,
        segmentation=False,
        normal=False
    ):
        self.simulation_frames = []

        self.res = res
        self.fps = fps
        self.period = 1.0 / fps
        self.segment_id = 0
        self.video_timestamp = time.time()
        self.current_frame = None

        self.frame_enabled = [rgb, depth, segmentation, normal]

    def stop(self):
        if len(self.simulation_frames) > 0:
            self.flush_frames()
        
        self.queue.put(None) # add poison pill

    def handle_frame(self, frame):
        timestamp = time.time()

        rgb_arr, depth_arr, seg_arr, normal_arr = frame

        if depth_arr is not None:
           depth_arr = VideoStreams.logarithmic_depth_to_rgb(depth_arr)

        simulation_frame = SimulationFrame(timestamp, rgb_arr, depth_arr, seg_arr, normal_arr)

        self.simulation_frames.append(simulation_frame)
        if self.video_timestamp + self.period <= simulation_frame.timestamp:
            self.flush_frames()

    def flush_frames(self):
        simulation_frame_paths = self.transcode_frames()
        self.queue.put(simulation_frame_paths)

    def transcode_frames(self) -> SimulationFramePaths:
        date_time = time.strftime('%Y%m%d_%H%M%S')
        if self.current_frame is None:
            self.current_frame = self.simulation_frames[0]

        for simulation_frame in self.simulation_frames:
            while self.video_timestamp + self.period < simulation_frame.timestamp:
                self.video_timestamp += self.period

            self.current_frame = simulation_frame

        simulation_frame_images = []
        for frame_id in range(len(VideoStreams.FRAME_TYPES)):
            if not self.frame_enabled[frame_id]:
                next

            filename = self._filename(VideoStreams.FRAME_TYPES[frame_id], date_time, self.segment_id)
            self.create_image_paths(self.current_frame.frame(frame_id), filename)
            simulation_frame_images.append(filename)

        self.simulation_frames.clear()
        self.segment_id += 1

        return SimulationFramePaths(self.current_frame.timestamp, simulation_frame_images)

    def create_image_paths(self, typed_array, filename):
        image = Image.fromarray(typed_array, mode='RGB')
        image.save(filename)
        return filename

    def _filename(self, frame_type, date_time, segment_id):
        return f"{Configuration.WORK_DIR}/{frame_type}_{date_time}_{segment_id}.webp"

